@@ -3,9 +3,9 @@
  *
  * Copyright 2000 by PostgreSQL Global Development Group
  *
- * $Header: /home/projects/pgsql/cvsroot/pgsql/src/bin/psql/describe.c,v 1.21 2000/04/16 20:04:51 petere Exp $
+ * $Header: /home/projects/pgsql/cvsroot/pgsql/src/bin/psql/describe.c,v 1.28 2001/03/22 04:00:19 momjian Exp $
  */
-#include "postgres.h"
+#include "postgres_fe.h"
 #include "describe.h"
 
 #include "libpq-fe.h"
@@ -44,10 +44,10 @@ describeAggregates(const char *name)
 	 * types ones that work on all
 	 */
 	strcpy(buf,
-		   "SELECT a.aggname AS \"Name\", t.typname AS \"Type\",\n"
+		   "SELECT a.aggname AS \"Name\", format_type(a.aggbasetype, NULL) AS \"Type\",\n"
 		   "  obj_description(a.oid) as \"Description\"\n"
-		   "FROM pg_aggregate a, pg_type t\n"
-		   "WHERE a.aggbasetype = t.oid\n"
+		   "FROM pg_aggregate a\n"
+		   "WHERE a.aggbasetype <> 0\n"
 		);
 
 	if (name)
@@ -103,7 +103,7 @@ describeFunctions(const char *name, bool verbose)
 	 * arguments, but have no types defined for those arguments
 	 */
 	strcpy(buf,
-		   "SELECT t.typname as \"Result\", p.proname as \"Function\",\n"
+		   "SELECT format_type(p.prorettype, NULL) as \"Result\", p.proname as \"Function\",\n"
 		   "       oidvectortypes(p.proargtypes) as \"Arguments\"");
 	if (verbose)
 		strcat(buf, ",\n       u.usename as \"Owner\", l.lanname as \"Language\", p.prosrc as \"Source\",\n"
@@ -111,13 +111,13 @@ describeFunctions(const char *name, bool verbose)
 
 	if (!verbose)
 		strcat(buf,
-			   "\nFROM pg_proc p, pg_type t\n"
-			   "WHERE p.prorettype = t.oid and (pronargs = 0 or oidvectortypes(p.proargtypes) != '')\n");
+			   "\nFROM pg_proc p\n"
+			   "WHERE p.prorettype <> 0 and (pronargs = 0 or oidvectortypes(p.proargtypes) <> '')\n");
 	else
 		strcat(buf,
-			   "\nFROM pg_proc p, pg_type t, pg_language l, pg_user u\n"
-			   "WHERE p.prorettype = t.oid AND p.prolang = l.oid AND p.proowner = u.usesysid\n"
-		"  AND (pronargs = 0 or oidvectortypes(p.proargtypes) != '')\n");
+			   "\nFROM pg_proc p,  pg_language l, pg_user u\n"
+			   "WHERE p.prolang = l.oid AND p.proowner = u.usesysid\n"
+			   "  AND p.prorettype <> 0 and (pronargs = 0 or oidvectortypes(p.proargtypes) <> '')\n");
 
 	if (name)
 	{
@@ -149,13 +149,16 @@ describeFunctions(const char *name, bool verbose)
 bool
 describeTypes(const char *name, bool verbose)
 {
-	char		buf[256 + REGEXP_CUTOFF];
+	char		buf[384 + 2 * REGEXP_CUTOFF];
 	PGresult   *res;
 	printQueryOpt myopt = pset.popt;
 
-	strcpy(buf, "SELECT t.typname AS \"Type\"");
+	strcpy(buf, "SELECT format_type(t.oid, NULL) AS \"Type\"");
 	if (verbose)
+	{
+		strcat(buf, ",\n  t.typname AS \"Internal name\"");
 		strcat(buf, ",\n  (CASE WHEN t.typlen = -1 THEN 'var'::text ELSE t.typlen::text END) as \"Size\"");
+	}
 	strcat(buf, ",\n  obj_description(t.oid) as \"Description\"");
 
 	/*
@@ -166,11 +169,14 @@ describeTypes(const char *name, bool verbose)
 
 	if (name)
 	{
-		strcat(buf, "  AND t.typname ~ '^");
+		/* accept either internal or external type name */
+		strcat(buf, "  AND (format_type(t.oid, NULL) ~ '^");
 		strncat(buf, name, REGEXP_CUTOFF);
-		strcat(buf, "' ");
+		strcat(buf, "' OR t.typname ~ '^");
+		strncat(buf, name, REGEXP_CUTOFF);
+		strcat(buf, "')");
 	}
-	strcat(buf, "ORDER BY t.typname;");
+	strcat(buf, "\nORDER BY \"Type\";");
 
 	res = PSQLexec(buf);
 	if (!res)
@@ -198,18 +204,13 @@ describeOperators(const char *name)
 
 	strcpy(buf,
 		   "SELECT o.oprname AS \"Op\",\n"
-		   "       t1.typname AS \"Left arg\",\n"
-		   "       t2.typname AS \"Right arg\",\n"
-		   "       t0.typname AS \"Result\",\n"
+		   "       format_type(o.oprleft, NULL) AS \"Left arg\",\n"
+		   "       format_type(o.oprright, NULL) AS \"Right arg\",\n"
+		   "       format_type(p.prorettype, NULL) AS \"Result\",\n"
 		   "       obj_description(p.oid) as \"Description\"\n"
-		   "FROM   pg_proc p, pg_type t0,\n"
-		   "       pg_type t1, pg_type t2,\n"
-		   "       pg_operator o\n"
-		   "WHERE  p.prorettype = t0.oid AND\n"
-		   "       RegprocToOid(o.oprcode) = p.oid AND\n"
-		   "       p.pronargs = 2 AND\n"
-		   "       o.oprleft = t1.oid AND\n"
-		   "       o.oprright = t2.oid\n");
+		   "FROM   pg_proc p, pg_operator o\n"
+		   "WHERE  RegprocToOid(o.oprcode) = p.oid AND\n"
+		   "       p.pronargs = 2\n");
 	if (name)
 	{
 		strcat(buf, "  AND o.oprname = '");
@@ -220,14 +221,12 @@ describeOperators(const char *name)
 	strcat(buf, "\nUNION\n\n"
 		   "SELECT o.oprname as \"Op\",\n"
 		   "       ''::name AS \"Left arg\",\n"
-		   "       t1.typname AS \"Right arg\",\n"
-		   "       t0.typname AS \"Result\",\n"
+		   "       format_type(o.oprright, NULL) AS \"Right arg\",\n"
+		   "       format_type(o.oprresult, NULL) AS \"Result\",\n"
 		   "       obj_description(p.oid) as \"Description\"\n"
-		   "FROM   pg_operator o, pg_proc p, pg_type t0, pg_type t1\n"
+		   "FROM   pg_operator o, pg_proc p\n"
 		   "WHERE  RegprocToOid(o.oprcode) = p.oid AND\n"
-		   "       o.oprresult = t0.oid AND\n"
-		   "       o.oprkind = 'l' AND\n"
-		   "       o.oprright = t1.oid\n");
+		   "       o.oprkind = 'l'\n");
 	if (name)
 	{
 		strcat(buf, "AND o.oprname = '");
@@ -237,15 +236,13 @@ describeOperators(const char *name)
 
 	strcat(buf, "\nUNION\n\n"
 		   "SELECT o.oprname  as \"Op\",\n"
-		   "       t1.typname AS \"Left arg\",\n"
+		   "       format_type(o.oprleft, NULL) AS \"Left arg\",\n"
 		   "       ''::name AS \"Right arg\",\n"
-		   "       t0.typname AS \"Result\",\n"
+		   "       format_type(o.oprresult, NULL) AS \"Result\",\n"
 		   "       obj_description(p.oid) as \"Description\"\n"
-		   "FROM   pg_operator o, pg_proc p, pg_type t0, pg_type t1\n"
+		   "FROM   pg_operator o, pg_proc p\n"
 		   "WHERE  RegprocToOid(o.oprcode) = p.oid AND\n"
-		   "       o.oprresult = t0.oid AND\n"
-		   "       o.oprkind = 'r' AND\n"
-		   "       o.oprleft = t1.oid\n");
+		   "       o.oprkind = 'r'\n");
 	if (name)
 	{
 		strcat(buf, "AND o.oprname = '");
@@ -339,7 +336,7 @@ permissionsList(const char *name)
 	strcat(descbuf, "SELECT relname as \"Relation\",\n"
 		   "       relacl as \"Access permissions\"\n"
 		   "FROM   pg_class\n"
-		   "WHERE  ( relkind = 'r' OR relkind = 'S') AND\n"
+		   "WHERE  relkind in ('r', 'v', 'S') AND\n"
 		   "       relname !~ '^pg_'\n");
 	if (name)
 	{
@@ -383,7 +380,7 @@ objectDescription(const char *object)
 	descbuf[0] = '\0';
 
 	/* Aggregate descriptions */
-	strcat(descbuf, "SELECT DISTINCT a.aggname as \"Name\", 'aggregate'::text as \"Object\", d.description as \"Description\"\n"
+	strcat(descbuf, "SELECT DISTINCT a.aggname::text as \"Name\", 'aggregate'::text as \"Object\", d.description as \"Description\"\n"
 		   "FROM pg_aggregate a, pg_description d\n"
 		   "WHERE a.oid = d.objoid\n");
 	if (object)
@@ -395,7 +392,7 @@ objectDescription(const char *object)
 
 	/* Function descriptions (except in/outs for datatypes) */
 	strcat(descbuf, "\nUNION ALL\n\n");
-	strcat(descbuf, "SELECT DISTINCT p.proname as \"Name\", 'function'::text as \"Object\", d.description as \"Description\"\n"
+	strcat(descbuf, "SELECT DISTINCT p.proname::text as \"Name\", 'function'::text as \"Object\", d.description as \"Description\"\n"
 		   "FROM pg_proc p, pg_description d\n"
 		   "WHERE p.oid = d.objoid AND (p.pronargs = 0 or oidvectortypes(p.proargtypes) != '')\n");
 	if (object)
@@ -407,7 +404,7 @@ objectDescription(const char *object)
 
 	/* Operator descriptions */
 	strcat(descbuf, "\nUNION ALL\n\n");
-	strcat(descbuf, "SELECT DISTINCT o.oprname as \"Name\", 'operator'::text as \"Object\", d.description as \"Description\"\n"
+	strcat(descbuf, "SELECT DISTINCT o.oprname::text as \"Name\", 'operator'::text as \"Object\", d.description as \"Description\"\n"
 		   "FROM pg_operator o, pg_description d\n"
 	/* must get comment via associated function */
 		   "WHERE RegprocToOid(o.oprcode) = d.objoid\n");
@@ -420,7 +417,7 @@ objectDescription(const char *object)
 
 	/* Type description */
 	strcat(descbuf, "\nUNION ALL\n\n");
-	strcat(descbuf, "SELECT DISTINCT t.typname as \"Name\", 'type'::text as \"Object\", d.description as \"Description\"\n"
+	strcat(descbuf, "SELECT DISTINCT format_type(t.oid, NULL) as \"Name\", 'type'::text as \"Object\", d.description as \"Description\"\n"
 		   "FROM pg_type t, pg_description d\n"
 		   "WHERE t.oid = d.objoid\n");
 	if (object)
@@ -432,7 +429,7 @@ objectDescription(const char *object)
 
 	/* Relation (tables, views, indices, sequences) descriptions */
 	strcat(descbuf, "\nUNION ALL\n\n");
-	strcat(descbuf, "SELECT DISTINCT c.relname as \"Name\", 'relation'::text||'('||c.relkind||')' as \"Object\", d.description as \"Description\"\n"
+	strcat(descbuf, "SELECT DISTINCT c.relname::text as \"Name\", 'relation'::text||'('||c.relkind||')' as \"Object\", d.description as \"Description\"\n"
 		   "FROM pg_class c, pg_description d\n"
 		   "WHERE c.oid = d.objoid\n");
 	if (object)
@@ -444,7 +441,7 @@ objectDescription(const char *object)
 
 	/* Rule description (ignore rules for views) */
 	strcat(descbuf, "\nUNION ALL\n\n");
-	strcat(descbuf, "SELECT DISTINCT r.rulename as \"Name\", 'rule'::text as \"Object\", d.description as \"Description\"\n"
+	strcat(descbuf, "SELECT DISTINCT r.rulename::text as \"Name\", 'rule'::text as \"Object\", d.description as \"Description\"\n"
 		   "FROM pg_rewrite r, pg_description d\n"
 		   "WHERE r.oid = d.objoid AND r.rulename !~ '^_RET'\n");
 	if (object)
@@ -456,7 +453,7 @@ objectDescription(const char *object)
 
 	/* Trigger description */
 	strcat(descbuf, "\nUNION ALL\n\n");
-	strcat(descbuf, "SELECT DISTINCT t.tgname as \"Name\", 'trigger'::text as \"Object\", d.description as \"Description\"\n"
+	strcat(descbuf, "SELECT DISTINCT t.tgname::text as \"Name\", 'trigger'::text as \"Object\", d.description as \"Description\"\n"
 		   "FROM pg_trigger t, pg_description d\n"
 		   "WHERE t.oid = d.objoid\n");
 	if (object)
@@ -573,7 +570,7 @@ describeTableDetails(const char *name, bool desc)
 	headers[1] = "Type";
 	cols = 2;
 
-	if (tableinfo.relkind == 'r')
+	if (tableinfo.relkind == 'r' || tableinfo.relkind == 'v')
 	{
 		cols++;
 		headers[cols - 1] = "Modifier";
@@ -589,13 +586,13 @@ describeTableDetails(const char *name, bool desc)
 
 
 	/* Get column info */
-	strcpy(buf, "SELECT a.attname, t.typname, a.attlen, a.atttypmod, a.attnotnull, a.atthasdef, a.attnum");
+	strcpy(buf, "SELECT a.attname, format_type(a.atttypid, a.atttypmod), a.attnotnull, a.atthasdef, a.attnum");
 	if (desc)
 		strcat(buf, ", obj_description(a.oid)");
-	strcat(buf, "\nFROM pg_class c, pg_attribute a, pg_type t\n"
+	strcat(buf, "\nFROM pg_class c, pg_attribute a\n"
 		   "WHERE c.relname = '");
 	strncat(buf, name, NAMEDATALEN);
-	strcat(buf, "'\n  AND a.attnum > 0 AND a.attrelid = c.oid AND a.atttypid = t.oid\n"
+	strcat(buf, "'\n  AND a.attnum > 0 AND a.attrelid = c.oid\n"
 		   "ORDER BY a.attnum");
 
 	res = PSQLexec(buf);
@@ -628,79 +625,30 @@ describeTableDetails(const char *name, bool desc)
 
 	for (i = 0; i < PQntuples(res); i++)
 	{
-		int4		attypmod = atoi(PQgetvalue(res, i, 3));
-		const char *attype = PQgetvalue(res, i, 1);
-		const char *typename;
-		bool		isarray = false;
-
 		/* Name */
 		cells[i * cols + 0] = PQgetvalue(res, i, 0);	/* don't free this
 														 * afterwards */
-
 		/* Type */
-		if (attype[0] == '_')
-		{
-			isarray = true;
-			attype++;
-		}
-		/* (convert some internal type names to SQL'ish) */
-		if (strcmp(attype, "int2") == 0)
-			typename = "smallint";
-		else if (strcmp(attype, "int4") == 0)
-			typename = "integer";
-		else if (strcmp(attype, "int8") == 0)
-			typename = "bigint";
-		else if (strcmp(attype, "bool") == 0)
-			typename = "boolean";
-		else
-			typename = attype;
-		/* more might need to be added when date/time types are sorted out */
-
-		cells[i * cols + 1] = xmalloc(NAMEDATALEN + 16);
-		if (strcmp(typename, "bpchar") == 0)
-		{
-			if (attypmod != -1)
-				sprintf(cells[i * cols + 1], "char(%d)", attypmod - VARHDRSZ);
-			else
-				sprintf(cells[i * cols + 1], "char()");
-		}
-		else if (strcmp(typename, "varchar") == 0)
-		{
-			if (attypmod != -1)
-				sprintf(cells[i * cols + 1], "varchar(%d)", attypmod - VARHDRSZ);
-			else
-				sprintf(cells[i * cols + 1], "varchar()");
-		}
-		else if (strcmp(typename, "numeric") == 0)
-		{
-			sprintf(cells[i * cols + 1], "numeric(%d,%d)",
-					((attypmod - VARHDRSZ) >> 16) & 0xffff,
-					(attypmod - VARHDRSZ) & 0xffff);
-		}
-		else
-			strcpy(cells[i * cols + 1], typename);
-
-		if (isarray)
-			strcat(cells[i * cols + 1], "[]");
-
+		cells[i * cols + 1] = PQgetvalue(res, i, 1);	/* don't free this
+														 * either */
 
 		/* Extra: not null and default */
 		/* (I'm cutting off the 'default' string at 128) */
-		if (tableinfo.relkind == 'r')
+		if (tableinfo.relkind == 'r' || tableinfo.relkind == 'v')
 		{
 			cells[i * cols + 2] = xmalloc(128 + 128);
 			cells[i * cols + 2][0] = '\0';
-			if (strcmp(PQgetvalue(res, i, 4), "t") == 0)
+			if (strcmp(PQgetvalue(res, i, 2), "t") == 0)
 				strcat(cells[i * cols + 2], "not null");
 
 			/* handle "default" here */
-			if (strcmp(PQgetvalue(res, i, 5), "t") == 0)
+			if (strcmp(PQgetvalue(res, i, 3), "t") == 0)
 			{
 				PGresult   *result;
 
 				sprintf(buf, "SELECT substring(d.adsrc for 128) FROM pg_attrdef d, pg_class c\n"
 						"WHERE c.relname = '%s' AND c.oid = d.adrelid AND d.adnum = %s",
-						name, PQgetvalue(res, i, 6));
+						name, PQgetvalue(res, i, 4));
 
 				result = PSQLexec(buf);
 				if (!result)
@@ -721,7 +669,7 @@ describeTableDetails(const char *name, bool desc)
 
 		/* Description */
 		if (desc)
-			cells[i * cols + cols - 1] = PQgetvalue(res, i, 7);
+			cells[i * cols + cols - 1] = PQgetvalue(res, i, 5);
 	}
 
 	/* Make title */
@@ -729,10 +677,10 @@ describeTableDetails(const char *name, bool desc)
 	switch (tableinfo.relkind)
 	{
 		case 'r':
-			if (view_def)
-				sprintf(title, "View \"%s\"", name);
-			else
-				sprintf(title, "Table \"%s\"", name);
+			sprintf(title, "Table \"%s\"", name);
+			break;
+		case 'v':
+			sprintf(title, "View \"%s\"", name);
 			break;
 		case 'S':
 			sprintf(title, "Sequence \"%s\"", name);
@@ -744,7 +692,8 @@ describeTableDetails(const char *name, bool desc)
 			sprintf(title, "Special relation \"%s\"", name);
 			break;
 		default:
-			sprintf(title, "?%c?", tableinfo.relkind);
+			sprintf(title, "?%c? \"%s\"", tableinfo.relkind, name);
+			break;
 	}
 
 	/* Make footers */
@@ -775,7 +724,7 @@ describeTableDetails(const char *name, bool desc)
 		}
 	}
 	/* Information about the view */
-	else if (tableinfo.relkind == 'r' && view_def)
+	else if (view_def)
 	{
 		footers = xmalloc(2 * sizeof(*footers));
 		footers[0] = xmalloc(20 + strlen(view_def));
@@ -926,8 +875,7 @@ describeTableDetails(const char *name, bool desc)
 
 	for (i = 0; i < PQntuples(res); i++)
 	{
-		free(cells[i * cols + 1]);
-		if (tableinfo.relkind == 'r')
+		if (tableinfo.relkind == 'r' || tableinfo.relkind == 'v')
 			free(cells[i * cols + 2]);
 	}
 	free(cells);
@@ -986,8 +934,7 @@ listTables(const char *infotype, const char *name, bool desc)
 		if (desc)
 			strcat(buf, ", obj_description(c.oid) as \"Description\"");
 		strcat(buf, "\nFROM pg_class c, pg_user u\n"
-			   "WHERE c.relowner = u.usesysid AND c.relkind = 'r'\n"
-			   "  AND not exists (select 1 from pg_views where viewname = c.relname)\n");
+			   "WHERE c.relowner = u.usesysid AND c.relkind = 'r'\n");
 		strcat(buf, showSystem ? "  AND c.relname ~ '^pg_'\n" : "  AND c.relname !~ '^pg_'\n");
 		if (name)
 		{
@@ -1002,7 +949,6 @@ listTables(const char *infotype, const char *name, bool desc)
 			strcat(buf, ", obj_description(c.oid) as \"Description\"");
 		strcat(buf, "\nFROM pg_class c\n"
 			   "WHERE c.relkind = 'r'\n"
-			   "  AND not exists (select 1 from pg_views where viewname = c.relname)\n"
 			   "  AND not exists (select 1 from pg_user where usesysid = c.relowner)\n");
 		strcat(buf, showSystem ? "  AND c.relname ~ '^pg_'\n" : "  AND c.relname !~ '^pg_'\n");
 		if (name)
@@ -1023,8 +969,7 @@ listTables(const char *infotype, const char *name, bool desc)
 		if (desc)
 			strcat(buf, ", obj_description(c.oid) as \"Description\"");
 		strcat(buf, "\nFROM pg_class c, pg_user u\n"
-			   "WHERE c.relowner = u.usesysid AND c.relkind = 'r'\n"
-			   "  AND exists (select 1 from pg_views where viewname = c.relname)\n");
+			   "WHERE c.relowner = u.usesysid AND c.relkind = 'v'\n");
 		strcat(buf, showSystem ? "  AND c.relname ~ '^pg_'\n" : "  AND c.relname !~ '^pg_'\n");
 		if (name)
 		{
@@ -1038,8 +983,7 @@ listTables(const char *infotype, const char *name, bool desc)
 		if (desc)
 			strcat(buf, ", obj_description(c.oid) as \"Description\"");
 		strcat(buf, "\nFROM pg_class c\n"
-			   "WHERE c.relkind = 'r'\n"
-			   "  AND exists (select 1 from pg_views where viewname = c.relname)\n"
+			   "WHERE c.relkind = 'v'\n"
 			   "  AND not exists (select 1 from pg_user where usesysid = c.relowner)\n");
 		strcat(buf, showSystem ? "  AND c.relname ~ '^pg_'\n" : "  AND c.relname !~ '^pg_'\n");
 		if (name)
@@ -1073,10 +1017,6 @@ listTables(const char *infotype, const char *name, bool desc)
 			strcat(buf, "'S'");
 		strcat(buf, ")\n");
 
-		/* ignore large-obj indices */
-		if (showIndices)
-			strcat(buf, "  AND (c.relkind != 'i' OR c.relname !~ '^xinx')\n");
-
 		strcat(buf, showSystem ? "  AND c.relname ~ '^pg_'\n" : "  AND c.relname !~ '^pg_'\n");
 		if (name)
 		{
@@ -1103,10 +1043,6 @@ listTables(const char *infotype, const char *name, bool desc)
 			strcat(buf, "'S'");
 		strcat(buf, ")\n");
 
-		/* ignore large-obj indices */
-		if (showIndices)
-			strcat(buf, "  AND (c.relkind != 'i' OR c.relname !~ '^xinx')\n");
-
 		strcat(buf, showSystem ? "  AND c.relname ~ '^pg_'\n" : "  AND c.relname !~ '^pg_'\n");
 		if (name)
 		{
